.chanlist {
	border           : 1px solid #1f669b;
	height			: 150px;
	overflow		: auto;
	background-color : #f1f1f1;
	width			: 600;
}

.chantable {
	border           : 0px;
	background-color : #f1f1f1;
	width			: 100%;
}

.labels {
	background-color : #000000;
	color : #ffffff;
}

.chanlisteven {
	background-color : #fff8e4;
}

.chanlistodd {
	background-color : #f0f5ff;
}	

.chanlistselected {
	background-color : #ffb13d;
}

.light {
	color : #717171;
}
