#
# Table structure for table `iaxfriends`
#

CREATE TABLE `iaxfriends` (
  `name` varchar(40) NOT NULL default '',
  `username` varchar(40) NOT NULL default '',
  `secret` varchar(40) NOT NULL default '',
  `dbsecret` varchar(40) NOT NULL default '',
  `context` varchar(40) NOT NULL default '',
  `regcontext` varchar(40) NOT NULL default '',
  `host` varchar(40) NOT NULL default 'dynamic',
  `ipaddr` varchar(20) NOT NULL default '',
  `port` int(6) NOT NULL default '0',
  `defaultip` varchar(20) NOT NULL default '',
  `sourceaddress` varchar(20) NOT NULL default '',
  `mask` varchar(20) NOT NULL default '',
  `regexten` varchar(40) NOT NULL default '',
  `regseconds` int(11) NOT NULL default '0',
  `accountcode` varchar(20) NOT NULL default '', 
  `mohinterpret` varchar(20) NOT NULL default '', 
  `mohsuggest` varchar(20) NOT NULL default '', 
  `inkeys` varchar(40) NOT NULL default '', 
  `outkey` varchar(40) NOT NULL default '', 
  `language` varchar(10) NOT NULL default '', 
  `callerid` varchar(40) NOT NULL default '', 
  `cid_number` varchar(40) NOT NULL default '', 
  `sendani` varchar(10) NOT NULL default '', 
  `fullname` varchar(40) NOT NULL default '', 
  `trunk` varchar(10) NOT NULL default '', 
  `auth` varchar(20) NOT NULL default '', 
  `maxauthreq` varchar(15) NOT NULL default '', 
  `encryption` varchar(20) NOT NULL default '', 
  `transfer` varchar(10) NOT NULL default '', 
  `jitterbuffer` varchar(10) NOT NULL default '', 
  `forcejitterbuffer` varchar(10) NOT NULL default '', 
  `disallow` varchar(40) NOT NULL default 'all', 
  `allow` varchar(40) NOT NULL default '', 
  `codecpriority` varchar(40) NOT NULL default '', 
  `qualify` varchar(10) NOT NULL default '', 
  `qualifysmoothing` varchar(10) NOT NULL default '', 
  `qualifyfreqok` varchar(10) NOT NULL default '', 
  `qualifyfreqnotok` varchar(10) NOT NULL default '', 
  `timezone` varchar(20) NOT NULL default '', 
  `adsi` varchar(10) NOT NULL default '', 
  `amaflags` varchar(20) NOT NULL default '', 
  `setvar` varchar(200) NOT NULL default '', 
  PRIMARY KEY  (`name`),
  INDEX name (name, host),
  INDEX name2 (name, ipaddr, port),
  INDEX ipaddr (ipaddr, port),
  INDEX host (host, port),
) TYPE=MyISAM;

