/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! 
 * \file
 * \brief Open Settlement Protocol (OSP)
 */

#ifndef _ASTERISK_OSP_H
#define _ASTERISK_OSP_H

#define AST_OSP_SUCCESS		((char*)"SUCCESS")		/* Return status, success */
#define AST_OSP_FAILED		((char*)"FAILED")		/* Return status, failed */
#define AST_OSP_ERROR		((char*)"ERROR")		/* Return status, error */

#endif /* _ASTERISK_OSP_H */
