/*! \file
  * \brief Random Data data
  *
  * Source: speex.raw
  *
  * Copyright (C) 1999-2005, Digium Inc.
  *
  * Distributed under the terms of the GNU General Public License
  *
  */

static unsigned char speex_slin_ex[] = {
0x2e, 0x8e, 0x0f, 0x9a, 0x20, 0000, 0x01, 0x7f, 0xff, 0xff, 
0xff, 0xff, 0xff, 0x91, 0000, 0xbf, 0xff, 0xff, 0xff, 0xff, 
0xff, 0xdc, 0x80, 0x5f, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
0x98, 0x7f, 0xff, 0xff, 0xff, 0xe8, 0xff, 0xf7, 0x80 };
