/*! \file
  * \brief Raw 8-bit data
  *
  * Source: ilbc.out
  *
  * Copyright (C) 1999-2005, Digium Inc.
  *
  * Distributed under the terms of the GNU General Public License
  *
  */

static unsigned char ilbc_slin_ex[] = {
0xff, 0xa0, 0xff, 0xfa, 0xf, 0x60, 0x12, 0x11, 0xa2, 0x47, 
0x22, 0x8c, 00, 00, 0x1, 0x2, 0x80, 0x43, 0xa0, 0x40, 
0x33, 0xff, 0xcf, 0xc0, 0xf3, 0xf3, 0x3f, 0x8f, 0x3f, 0xff, 
0xff, 0xff, 0xff, 0xfc, 0xf9, 0xe5, 0x55, 0x78, 0xb, 0xca, 
0xe1, 0x27, 0x94, 0x7b, 0xa8, 0x91, 0x2c, 0x36, 0x8, 0x56 };
